/*

  DBOX Module
  Copyright (C) 1994-2000 Daniel Kroening <kroening@handshake.de>

  Purpose: main file of check; tests ZCONNECT and ERFC files

*/

#include <string.h>
#include <ctype.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>
#include <time.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/mman.h>

#include <config.h>
#include <mtoolzc.h>
#include <mtoolerfc.h>
#include <tools.h>
#include <dboxversion.h>
#include <libdboxversion.h>

void printmessage(zcmessaget *m)
 {
  char tempstr[TEXTLEN];

  m->copyheadertext(tempstr, m->mid, TEXTLEN-1);
  cout << "LEN: " << m->len << " MID: " << tempstr << "\n";
 }

void printmessage(erfcmessaget *m)
 {
  //char tempstr[TEXTLEN];

  //m->copyheadertext(tempstr, m->mid, TEXTLEN-1);
  //cout << "MID: " << tempstr << "\n";

  cout << "LEN: " << m->len << "\n";
 }

bool check_zc_file(const char *messagefile, dword length)
 {
  bool error=FALSE;

  zcmessaget m;
  m.data=messagefile;
  m.endoffile=messagefile+length;

  while(!m.eof() && !error)
   {
    if(m.parse() || m.mid==NULL)
      error=TRUE;
    else
     {
      printmessage(&m);
      m.data=m.endofmail;
     }
   }

  return error;
 }

bool check_erfc_file(const char *messagefile, dword length)
 {
  bool error=FALSE;

  erfcmessaget m;
  m.data=messagefile;
  m.endoffile=messagefile+length;

  while(!m.eof() && !error)
   {
    if(m.parse())
      error=TRUE;
    else
     {
      printmessage(&m);
      m.data=m.endofmail;
     }
   }

  return error;
 }

bool check_file(const char *messagefile, dword length)
 {
  const char *tptr=messagefile;
  dword i;
  enum { NONE, ZC, ERFC } type=NONE;

  for(i=0; i<length; i++, tptr++)
   {
    if(*tptr=='\n') { type=ERFC; break; }
    if(*tptr=='\r') { type=ZC; break; }
   }

  switch(type)
   {
   case ZC:
    cout << "File type: ZCONNECT\n\n";
    return check_zc_file(messagefile, length);

   case ERFC:
    cout << "File type: ERFC\n\n";
    return check_erfc_file(messagefile, length);
   
   default:
    cerr << "UNKNOWN FILE TYPE\n";
   }

  return TRUE;
 }

void check_file(const char *filename)
 {
  int handle;
  dword length;
  char *messagefile=(char *)NULL;
  bool error=FALSE, mapped=FALSE;

  chdir(config.cwd);
  handle=open(filename, O_RDONLY|O_BINARY);
  config.dboxcd();

  if(handle<0) return;

  length=filelength(handle);

  cout << "File size: " << length << " Bytes\n";

  if(length==0) return;

  if(length>60000)
   {
    messagefile=(char *)mmap(NULL, length, PROT_READ, MAP_FILE|MAP_PRIVATE, handle, 0);
    mapped=TRUE;
    if(messagefile==NULL)
     {
      cerr << "mmap() failed!\n";
      return;
     }
   }

  if(messagefile==NULL)
   {
    messagefile=(char *)malloc(length);
    if(messagefile==NULL)
     {
      cerr << "malloc() failed!\n";
      return;
     }

    if(read(handle, messagefile, length)!=(signed long)length)
     {
      cerr << "read() failed!\n";
      return;
     }
   }

  error=check_file(messagefile, length);

  if(mapped)
    munmap(messagefile, length);
  else
    free(messagefile);

  if(error)
    cout << "Message file has errors!\n";
  else
    cout << "Message file ok!\n";

  close(handle);
 }

static bool parseoptions(int argc, char **argv)
 {
  if(argc<2) goto help;
  if(argc>2) goto aufruf;

  if(argv[1][0]=='-') goto help;

  check_file(argv[1]);

  return FALSE;

 aufruf:
  cerr << "Falscher Aufruf!\n";

 help:
  cerr << 
    "\n"
    "* * CHECK - DBOX Module for message file checking - version "
    DBOXVERSIONNR " * *\n"
    "\n"
    "Usage:                       Effect:\n"
    "\n"
    " check [-?] [-h]              display help\n"
    " check filename               check message file\n"
    "\n"
    "Further options:\n"
    "\n";

  return TRUE;
 }

int main(int argc, char **argv)
 {
  if(init_libdbox(LIBDBOXVERSION, LIBDBOXVERSIONNR)) return 1;

  if(config.read("")) return 1;

  parseoptions(argc, argv);

  return 0;
 }
