#include <unistd.h>

#include <stringtools.h>
#include <dboxpath.h>
#include <zconnect.h>
#include <logf.h>
#include <messageid.h>
#include <spoolp.h>

#include "bouncezc.h"

void bounce(zcmessaget &m, 
            userpt &up,
            vector<bounce_infot> rcpts)
 {
  char from[TEXTLEN], mid[TEXTLEN], *tptr;

  string bounce_log;

  getboxlogdir(bounce_log, config.boxdir);
  bounce_log+="bounce.log";

  if(m.mid!=NULL)
    m.copyheadertext(mid, m.mid, TEXTLEN-1);
  else
    mid[0]=0;

  from[0]=0;

  if(m.wab!=NULL)
   {
    m.copyheadertext(from, m.wab, TEXTLEN-1);
    tptr=strchr(from, ' ');
    if(tptr!=NULL) *tptr=0;
   }

  if(from[0]==0 && m.abs!=NULL)
   {
    m.copyheadertext(from, m.abs, TEXTLEN-1);
    tptr=strchr(from, ' ');
    if(tptr!=NULL) *tptr=0;
   }

  logf.fileprintf(bounce_log.c_str(), "bounce",
    "mail from \"%s\" msgid \"%s\"", from, mid);
   
  for(vector<bounce_infot>::const_iterator it=rcpts.begin();
      it!=rcpts.end(); it++)
    logf.fileprintf(bounce_log.c_str(), "bounce", "  to %s... %s",
      it->rcpt.c_str(), it->errormsg.c_str());

   {
    char tempstr[TEXTLEN];
    strmaxcpy(tempstr, from, TEXTLEN-1);
    tptr=strchr(tempstr, '@');
    if(tptr==NULL)
     {
      logf.printf("bounce", "  invalid sender, bouncing futile");
      return;
     }

    if(tptr!=NULL) *tptr=0;

    if(stricmp(tempstr, "mailer-daemon")==0)
     {
      logf.printf("bounce", "  ignored!");
      return;
     }
   }

  string text;

  text=
    "The original message was received at (unknown)\r\n"
    "from unknown@"+string(up.name)+"\r\n"
    "\r\n"
    "   ----- The following addresses had delivery problems -----\r\n";

  for(vector<bounce_infot>::const_iterator it=rcpts.begin();
      it!=rcpts.end(); it++)
   {
    text+=it->rcpt;
    text+="\r\n";
   }

  text+=
    "\r\n"
    "   ----- Transcript of session follows -----\r\n";

  for(vector<bounce_infot>::const_iterator it=rcpts.begin();
      it!=rcpts.end(); it++)
   {
    text+="550 ";
    text+=it->rcpt;
    text+="... ";
    text+=it->errormsg;
    text+="\r\n";
   }

  text+=
    "\r\n"
    "   ----- Original message follows -----\r\n";

  string tmpfile;

  getdboxtempdir(tmpfile, config.dboxdir);
  tmpfile+="spool.bounce."+i2string(getpid());

  FILE *outfile=fopen(tmpfile.c_str(), "wb");
  if(outfile==NULL)
   {
    logf.printf("bounce", "failed to create \"%s\"", tmpfile.c_str());
    return;
   }

  fprintf(outfile, "ABS:\tMAILER-DAEMON@%s.%s\r\n",
    config.boxname, config.domain);

  fprintf(outfile, "BET:\tReturned mail: %s\r\n", rcpts[0].errormsg.c_str());

  fprintf(outfile, "LEN:\t%lu\r\n", text.size()+m.headerlen+m.len);

  fprintf(outfile, "EDA:\t%s\r\n", make_eda_localtime().c_str());

  fprintf(outfile, "EMP:\t%s\r\n", from);

  fprintf(outfile, "MAILER:\tDBOX spooler\r\n"); 

  fprintf(outfile, "MID:\t%s\r\n",
    gen_message_id("spool").c_str());

  fprintf(outfile, "ROT:\r\n");

  fprintf(outfile, "\r\n");

  fwrite(text.c_str(), 1, text.size(), outfile);
  fwrite(m.data, 1, m.headerlen, outfile);
  fwrite(m.data+m.headerlen, 1, m.len, outfile);

  fclose(outfile);

  if(spool_zc_puffer_mail(tmpfile.c_str(), config.boxname))
   {
    logf.printf("bounce", "spool_zc_puffer_mail failed");
    return;
   }
 }
