/*

  DBOX Module
  Copyright (C) 1994-2000 Daniel Kroening <kroening@handshake.de>

  Purpose: spool ERFC mail

*/

#include <unistd.h>
#include <dirent.h>
#include <string.h>
#include <ctype.h>
#include <fcntl.h>
#include <stdlib.h>
#include <stdarg.h>
#include <sys/stat.h>
#include <time.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/mman.h>

#include <dboxpath.h>
#include <tools.h>
#include <mtoolerfc.h>
#include <stool.h>
#include <lock.h>
#include <logf.h>
#include <move2bad.h>
#include <dboxversion.h>
#include "logmail.h"
#include "spool.h"
#include "spoolerfcm.h"
#include "sdisp.h"
#include "mroute.h"
#include "delivererfc.h"
#include "sptools.h"

class spoolerfcmt
 {
 public:
  struct destusert
   {
    char username[TEXTLEN];
    byte RCPTs[MAXEMPS];
   } *destusers;

  dword number_of_destusers;
  userpt up;
  erfcmessaget erfcm;
  string junk_errors;

  bool check_env_from();
  void docopymail();
  void parse_RCPT(const char *RCPT, string &host, string &user);
  void check_RCPTs(dword &number_of_RCPTs);
  void addemp(dword RCPT_nr, dword number_of_RCPTs, const string &dest);
  void doRCPT(dword RCPT_nr, dword number_of_RCPTs, const string &dest,
              const string &user, const string &host, const string &rcpt);
  void spoolmail(dword number_of_RCPTs);
  void spoolposting(dword pos);
  bool spoolfile(int handle);
  void spoolfile(const char *filename);
  void bounce();
  void posteb(const string &rcpt);

  spoolerfcmt();
  ~spoolerfcmt();
 };

spoolerfcmt::spoolerfcmt()
 {
 }

spoolerfcmt::~spoolerfcmt()
 {
 }

void spoolerfcmt::bounce()
 {
  if(junk_errors=="") return;


 }

bool spoolerfcmt::check_env_from()
 {
  string env_from;

  if(!erfcm.findsingleheader(env_from, "FROM"))
   {
    if(checkabscharset(env_from.c_str(), atoi(up|"system")))
      return TRUE;
   }

  return FALSE;
 }

void spoolerfcmt::docopymail()
 {
  string received=
    "(from "+string(up.name)+") "
    "by "+config.boxname+"."+config.domain+
    " (DBOX " DBOXVERSIONNR ")\n";

  for(dword nr=0; nr<number_of_destusers; nr++)
   {
    filet outfile;
    char filename[DIRLEN];

    if(!user_is_box(destusers[nr].username))
     {
      snprintf(filename, DIRLEN, BOXSPOOLDIR "%s.erfc", 
              config.boxdir, destusers[nr].username, 
              destusers[nr].username);

      display.versenden(config.boxname, destusers[nr].username, "Puffer");
     }
    else
     {
      char tmp_boxdir[DIRLEN];

      snprintf(tmp_boxdir, DIRLEN, DBOXBOXDIR, 
        config.dboxdir, destusers[nr].username);

      snprintf(filename, DIRLEN, BOXSPOOLINRFCMAILDIR "%s.erfc", 
              tmp_boxdir, config.boxname);

      display.versenden(destusers[nr].username, "", "inbound");
     }

    while(lock("puffer.prv", destusers[nr].username, config.boxname))
     {
      display.console.printf("%s is locked... ",
         destusers[nr].username);
      sleep(10);
     }

    if(outfile.open(filename, "ab"))
     {
      logf.printf("docopymail", "fatal: failed to open \"%s\"",
                  filename);
     }
    else
     {
      erfcm.copyheader_special(outfile.file, destusers[nr].RCPTs);
      fprintf(outfile.file, "%s", received.c_str());
      erfcm.copybody(outfile.file);
     }

    unlock("puffer.prv", destusers[nr].username, config.boxname);
   }
 }

void spoolerfcmt::check_RCPTs(dword &number_of_RCPTs)
 {
  const char *tptr, *text;

  number_of_RCPTs=0;

  tptr=erfcm.data;

  while(!erfcm.findheader(&tptr, "RCPT", &text))
    number_of_RCPTs++;
 }

void spoolerfcmt::addemp(dword RCPT_nr, dword number_of_RCPTs, 
                         const string &dest)
 {
  dword i, j;

  for(i=0; i<number_of_destusers; i++)
    if(strcmp(destusers[i].username, dest.c_str())==0)
      break;

  if(i==number_of_destusers)
   {
    number_of_destusers++;

    for(j=0; j<number_of_RCPTs; j++)
      destusers[i].RCPTs[j]=DOSKIP;
   }

  strmaxcpy(destusers[i].username, dest.c_str(), TEXTLEN-1);

  destusers[i].RCPTs[RCPT_nr]=DOEMP;
 }

void spoolerfcmt::posteb(const string &rcpt)
 {
  // TODO
 }

void spoolerfcmt::doRCPT(dword RCPT_nr, dword number_of_RCPTs,
                         const string &dest, const string &host,
                         const string &user, const string &rcpt)
 {
  bool pop=FALSE, puf=FALSE;
  char filename[DIRLEN], tempstr[TEXTLEN*2];
  userpt destup;

  if(stricmp(dest.c_str(), config.boxname)==0)
   {
    if(strchr(user.c_str(), '/')!=NULL)
     {
      junk_errors+=rcpt+": user is invalid\n";
      return;
     }

    strmaxcpy(destup.name, user.c_str(), TEXTLEN-1);
    dnstring(destup.name);

    if(destup.read())
     {
      junk_errors+=rcpt+": user is unknown\n";
      return;
     }

    if((destup|"alias")[0]!=0 &&
       !atoi(destup|"account"))
     {
      strmaxcpy(destup.name, destup|"alias", TEXTLEN-1);
      if(destup.read())
       {
        junk_errors+=rcpt+": alias "+string(up.name)+
                     " is unknown\n";
        return;
       }
     }

    destup.get("disablepop", tempstr);
    if(!atoi(tempstr)) pop=TRUE;

    snprintf(filename, DIRLEN, "%smaps", destup.getpath());
    if(access(filename, 00)==0 && !atoi(destup|"system"))
      puf=TRUE;
   }
  else
   {
    destup.setname(dest);
    if(destup.read())
     {
      junk_errors+=rcpt+": did not find dest \""+dest+"\"\n";
      return;
     }

    destup.get("disablepop", tempstr);
    if(!atoi(destup|"system") && !atoi(tempstr))
      pop=TRUE;

    snprintf(filename, DIRLEN, "%smaps", destup.getpath());
    if(access(filename, 00)==0 || atoi(destup|"system"))
      puf=TRUE;
   }

  if(pop)
   {
    display.versenden(config.boxname, destup.name, "mailbox");
    deliver_mail(erfcm, destup, rcpt);
    posteb(rcpt);
   }

  if(puf)
    addemp(RCPT_nr, number_of_RCPTs, destup.name);

  logmail(up.name, destup.name, erfcm, rcpt);
 }

void spoolerfcmt::parse_RCPT(const char *RCPT,
                             string &host, string &user)
 {
  user=host="";

  for(; *RCPT!=0; RCPT++)
   {
    if(*RCPT=='@')
     {
      RCPT++;
      host=RCPT;
      return;
     }

    user+=*RCPT;
   }
 }

void spoolerfcmt::spoolmail(dword number_of_RCPTs)
 {
  const char *dataptr, *text;
  dword count=0;

  dataptr=erfcm.data;
  while(!erfcm.findheader(&dataptr, "RCPT", &text) && count!=MAXEMPS)
   {
    char junkRCPT[TEXTLEN];
    string host, user;

     {
      char tempstr[1000];
      erfcm.copyheadertext(tempstr, text, 999);
      strmaxcpy(junkRCPT, tempstr, TEXTLEN-1);
      dnstring(tempstr);
      parse_RCPT(tempstr, host, user);
     }

    if(host=="" || strchr(host.c_str(), ' ')!=NULL ||
       strchr(user.c_str(), ' ')!=NULL ||
       strchr(host.c_str(), '.')==NULL)
     {
      junk_errors+=junkRCPT;
      junk_errors+=": invalid RCPT\n";
     }
    else
     {
      dword maxlen;
      string dest;

      if(mailroute(dest, up.name, host, maxlen))
       {
        junk_errors+=junkRCPT;
        junk_errors+=": no route to "+host+"\n";
       }
      else
       {
        if(erfcm.len>maxlen && maxlen!=0)
         {
          junk_errors+=junkRCPT;
          junk_errors+=": mail too large\n";
         }
        else
          doRCPT(count, number_of_RCPTs, dest, user, host, junkRCPT);
       }
     }

    count++;
   }

  if(count!=number_of_RCPTs)
    logf.internalerror("spoolnews", "count!=number_of_RCPTs");
 }

void spoolerfcmt::spoolposting(dword pos)
 {
  dword number_of_RCPTs;

  junk_errors="";

  display.startmail(config.boxname, up.name, &erfcm, pos);

  check_RCPTs(number_of_RCPTs);

  if(number_of_RCPTs>MAXEMPS)
   {
    junk_errors="Too many receipients\n";
    bounce();
    return;
   }

  if(number_of_RCPTs==0)
   {
    junk_errors="No receipients\n";
    bounce();
    return;
   }

   {
    char tempstr[1000];
    erfcm.findsingleheader(tempstr, "Message-Id", 999);
    if(tempstr[0]!=0 && checkmid(tempstr, atoi(up|"system")))
     {
      junk_errors="Invalid Message-Id";
      bounce();
      return;
     }
   }

  if(check_env_from())
   {
    junk_errors="Invalid Envelope FROM";
    bounce();
    return;
   }

  destusers=(destusert *)malloc(sizeof(destusert)*number_of_RCPTs);

  if(destusers==NULL)
    logf.internalerror("spoolposting", "FATAL: malloc failed");

  number_of_destusers=0;

  spoolmail(number_of_RCPTs);
  docopymail();

  free(destusers);

  bounce();
 }

bool spoolerfcmt::spoolfile(int handle)
 {
  dword length;
  char *posting=NULL;
  bool error=FALSE, mapped=FALSE;
  time_t t, lastcheck=0;

  length=filelength(handle);

  logf.printf("spoolfile", "Verteile %lu Bytes von System "
    "%s@%s", length, up.name, config.boxname);

  if(length<=2) return FALSE;

  if(length>60000)
   {
    posting=(char *)mmap(NULL, length, PROT_READ, MAP_FILE|MAP_PRIVATE, handle, 0);
    mapped=TRUE;
    if(posting==NULL)
     {
      logf.printf("spoolfile", "Fatal: mmap failed!");
      return TRUE;
     }
   }

  if(posting==NULL)
   {
    posting=(char *)malloc(length);
    if(posting==NULL) return TRUE;

    if(read(handle, posting, length)!=(ssize_t)length)
      return TRUE;
   }

  erfcm.data=posting;
  erfcm.endoffile=posting+length;

  while(!erfcm.eof() && !error)
   {
    if(display.checkabort())
     {
      error=TRUE;
      logf.printf("spoolfile", "spooling aborted manually");
     }
    else if((time(&t)-lastcheck)>10)
     {
      lastcheck=t;
      if(checkload(maxload))
       {
        if(display.waitload())
          maxload=0;
       }
     }
    else
     {
      if(erfcm.parse())
        error=TRUE;
      else
       {
        spoolposting(erfcm.data-posting);
        erfcm.data=erfcm.endofmail;
       }
     }
   }

  if(mapped)
    munmap(posting, length);
  else
    free(posting);

  return error;
 }

void spoolerfcmt::spoolfile(const char *filename)
 {
  char *tptr, tempstr[DIRLEN];
  int handle;
  bool error;

  strmaxcpy(tempstr, filename, DIRLEN-1);
  tptr=strrchr(tempstr, '.');
  *tptr=0;
  strmaxcpy(up.name, tempstr, TEXTLEN-1);

  if(up.read())
   {
    logf.printf("spoolfile", "fatal: unknown user %s@%s", 
      up.name, config.boxname);
    movetobad(TRUE, filename, config.boxdir);
    return;
   }

  snprintf(tempstr, DIRLEN, BOXSPOOLINRFCMAILDIR "%s", 
    config.boxdir, filename);

  handle=open(tempstr, O_RDONLY|O_BINARY);

  if(handle<0)
   {
    logf.printf("spoolfile", "fatal: failed to open \"%s\"", tempstr);
    movetobad(TRUE, filename, config.boxdir);
    return;
   }

  error=spoolfile(handle);

  close(handle);

  if(error)
    movetobad(TRUE, filename, config.boxdir);
  else
   {
    logf.printf("spoolfile", "spooling done");

    snprintf(tempstr, DIRLEN, BOXSPOOLINRFCMAILDIR "%s",
      config.boxdir, filename);

    if(remove(tempstr)!=0) 
      logf.printf("spoolfile", "remove(\"%s\") failed!", tempstr);
   }
 }

void getfilename(char *filename)
 {
  DIR *dir;
  struct dirent *ent;
  char tempstr[DIRLEN];

  filename[0]=0;

  snprintf(tempstr, DIRLEN, BOXSPOOLINRFCMAILDIR ".", config.boxdir);
  if((dir=opendir(tempstr))==NULL)
   {
    logf.printf("dospool", "Fatal: Failed to open SPOOLINDIR!");
    return;
   }

  while((ent=readdir(dir))!=NULL)
    if(ent->d_name[0]!='.' && ent->d_name[0]!='_')
     {
      if(strchr(ent->d_name, '.')!=NULL)
       {
        strmaxcpy(filename, ent->d_name, DIRLEN-1);
        closedir(dir);
        return;
       }
     }

  closedir(dir);
 }

void dospoolerfcm()
 {
  bool end=FALSE, did_something=FALSE;

  cout << "Starting Spooler\n";

  display.open();

  do
   {
    char filename[DIRLEN];

    getfilename(filename);
    if(filename[0]==0)
      end=TRUE;
    else
     {
      spoolerfcmt spoolerfcm;

      display.console.printf("\r\nSpooling \"%s\" from %s...\r\n", 
        filename, config.boxname);

      spoolerfcm.spoolfile(filename);

      did_something=TRUE;
     }
   }
  while(!end);

  display.close();

  cout << "\n";

  if(!did_something)
   {
    logf.timefprintf(stdout);
    cout << "Nothing to spool\n";
   }
 }

